/*
 *  audio_decoder_addon.cpp
 *  Copyright (C) 2005 Marcus Overhagen <marcus@overhagen.de>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "audio_decoder_addon.h"

#define TRACE_MP3_DECODER
#ifdef TRACE_MP3_DECODER
  #define TRACE printf
#else
  #define TRACE(a...)
#endif


#define DECODE_BUFFER_SIZE	(32 * 1024)


inline size_t
AudioBufferSize(const media_raw_audio_format &raf, bigtime_t buffer_duration = 50000 /* 50 ms */)
{
	return (raf.format & 0xf) * (raf.channel_count)
         * (size_t)((raf.frame_rate * buffer_duration) / 1000000.0);
}


// bit_rate_table[mpeg_version_index][layer_index][bitrate_index]
static const int bit_rate_table[4][4][16] =
{
	{ // mpeg version 2.5
		{ }, // undefined layer
		{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0 }, // layer 3
		{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0 }, // layer 2
		{ 0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, 0 } // layer 1
	},
	{ // undefined version
		{ },
		{ },
		{ },
		{ }
	},
	{ // mpeg version 2
		{ },
		{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0 },
		{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0 },
		{ 0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, 0 }
	},
	{ // mpeg version 1
		{ },
		{ 0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0 },
		{ 0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },
		{ 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0}
	}
};


// frame_rate_table[mpeg_version_index][sampling_rate_index]
static const int frame_rate_table[4][4] =
{
	{ 11025, 12000, 8000, 0},	// mpeg version 2.5
	{ 0, 0, 0, 0 },
	{ 22050, 24000, 16000, 0},	// mpeg version 2
	{ 44100, 48000, 32000, 0}	// mpeg version 1
};



AudioDecoder::AudioDecoder(get_next_chunk_func next_chunk, void *cookie)
 :	DecoderInterface(next_chunk, cookie)
{
	static BLocker initLocker;
	static bool initDone = false;
	initLocker.Lock();
	if (!initDone) {
		InitMpgLib();
		initDone = true;
	}	
	initLocker.Unlock();
	InitMP3(&fMpgLibPrivate);
	fResidualBytes = 0;
	fResidualBuffer = 0;
	fDecodeBuffer = new uint8 [DECODE_BUFFER_SIZE];
	fStartTime = 0;
	fFrameSize = 0;
	fFrameRate = 0;
	fBitRate = 0;
	fChannelCount = 0;
	fOutputBufferSize = 0;
	fNeedSync = false;

}


AudioDecoder::~AudioDecoder()
{
	ExitMP3(&fMpgLibPrivate);
	delete [] fDecodeBuffer;
}


status_t
AudioDecoder::SetInputFormat(const media_format &in_format)
{

	// decode first chunk to initialize mpeg library
	if (B_OK != DecodeNextChunk()) {
		printf("AudioDecoder::SetInputFormat failed, can't decode first chunk\n");
		return B_ERROR;
	}

	// initialize fBitRate, fFrameRate and fChannelCount from mpg decode library values of first header
	extern int tabsel_123[2][3][16];
	extern long freqs[9];
	fBitRate = tabsel_123[fMpgLibPrivate.fr.lsf][fMpgLibPrivate.fr.lay-1][fMpgLibPrivate.fr.bitrate_index] * 1000;
	fFrameRate = freqs[fMpgLibPrivate.fr.sampling_frequency];
	fChannelCount = fMpgLibPrivate.fr.stereo;
	
	printf("AudioDecoder::SetInputFormat: channels %d, bitrate %d, framerate %d\n", fChannelCount, fBitRate, fFrameRate);

	return B_OK;
}


status_t
AudioDecoder::SetOutputFormat(media_format *output_format)
{
	output_format->type = B_MEDIA_RAW_AUDIO;
	output_format->u.raw_audio.frame_rate = fFrameRate;
	output_format->u.raw_audio.channel_count = fChannelCount;
	output_format->u.raw_audio.format = media_raw_audio_format::B_AUDIO_SHORT;
	output_format->u.raw_audio.byte_order = B_MEDIA_HOST_ENDIAN;
	
	int frame_size = (output_format->u.raw_audio.format & 0xf) * output_format->u.raw_audio.channel_count;
	if (output_format->u.raw_audio.buffer_size == 0)
		output_format->u.raw_audio.buffer_size = AudioBufferSize(output_format->u.raw_audio);
	else
		output_format->u.raw_audio.buffer_size = (output_format->u.raw_audio.buffer_size / frame_size) * frame_size;
	if (output_format->u.raw_audio.channel_mask == 0)
		output_format->u.raw_audio.channel_mask = (fChannelCount == 1) ? B_CHANNEL_LEFT : B_CHANNEL_LEFT | B_CHANNEL_RIGHT;

	// setup rest of the needed variables
	fFrameSize = frame_size;
	fOutputBufferSize = output_format->u.raw_audio.buffer_size;
	
	return B_OK;
}


status_t
AudioDecoder::Decode(void *buffer, int64 *frameCount,
						 media_header *mediaHeader, media_decode_info *info)
{
	status_t last_err = B_LAST_BUFFER_ERROR;
	uint8 * out_buffer = static_cast<uint8 *>(buffer);
	int32	out_bytes_needed = fOutputBufferSize;
	int32	out_bytes = 0;
	media_header_ex *mhe = (media_header_ex *)mediaHeader;
	
	mediaHeader->start_time = fStartTime;
	//TRACE("mp3Decoder: Decoding start time %.6f\n", fStartTime / 1000000.0);
	
	while (out_bytes_needed > 0) {
		if (fResidualBytes) {
			int32 bytes = min_c(fResidualBytes, out_bytes_needed);
			memcpy(out_buffer, fResidualBuffer, bytes);
			fResidualBuffer += bytes;
			fResidualBytes -= bytes;
			out_buffer += bytes;
			out_bytes += bytes;
			out_bytes_needed -= bytes;
			
			fStartTime += (1000000LL * (bytes / fFrameSize)) / fFrameRate;

//			TRACE("mp3Decoder: fStartTime inc'd to %Ld\n", fStartTime);
			continue;
		}
		
		last_err = DecodeNextChunk();
		if (last_err != B_OK) {
			break;
		}
	}

	*frameCount = out_bytes / fFrameSize;

	mhe->type = B_MEDIA_RAW_AUDIO;
	mhe->size_used = *frameCount * fFrameSize;
	mhe->u.raw_audio.frame_rate = fFrameRate;
	mhe->u.raw_audio.channel_count = fChannelCount;

	return (out_bytes > 0) ? B_OK : last_err;
}


status_t
AudioDecoder::DecodeNextChunk()
{
	const void *chunkBuffer;
	size_t chunkSize;
	media_header mh;
	int outsize;
	int result;
	status_t err;

	// decode residual data that is still in the decoder first
	result = decodeMP3(&fMpgLibPrivate, 0, 0, (char *)fDecodeBuffer, DECODE_BUFFER_SIZE, &outsize);
	if (result == MP3_OK) {
		fResidualBuffer = fDecodeBuffer;
		fResidualBytes = outsize;
		return B_OK;
	}
	
//	printf("decode residual data result %d\n", result);
	
	// get another chunk and push it to the decoder
	
	int max_trys;
	for (max_trys = 50; max_trys > 0; max_trys--) {
		err = GetNextChunk(&chunkBuffer, &chunkSize, &mh);
		if (err != B_OK) {
			printf("AudioDecoder: GetNextChunk failed\n");
			return err;
		}
		if (!IsValidStream((const uint8 *)chunkBuffer, chunkSize)) {
			printf("AudioDecoder: lost sync\n");
		} else {
			break;
		}
	}
	if (max_trys == 0) {
		TRACE("AudioDecoder: Sync failed\n");
		return B_ERROR;
	}

	
	fStartTime = mh.start_time + 190000; // decoder delay offset ???
//	TRACE("mp3Decoder: fStartTime reset to %Ld\n", fStartTime);

//	printf("ready to decode %d bytes into a %d byte buffer\n", chunkSize, DECODE_BUFFER_SIZE);

	{
//		const uint8 *p = (const uint8 *)chunkBuffer;
//		printf("in %02x %02x %02x %02x %02x %02x %02x %02x\n", p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
	}

	result = decodeMP3(&fMpgLibPrivate, (char *)chunkBuffer, chunkSize, (char *)fDecodeBuffer, DECODE_BUFFER_SIZE, &outsize);
	if (result == MP3_ERR) {
		TRACE("mp3Decoder::Decode: decodeMP3 returned MP3_ERR\n");
		ExitMP3(&fMpgLibPrivate);
		InitMP3(&fMpgLibPrivate);
		printf("AudioDecoder: Reset after decoding error\n");
		fResidualBytes = 0;
		fResidualBuffer = 0;
		return B_ERROR;
	}

	if (freqs[fMpgLibPrivate.fr.sampling_frequency] != 0) {
		fFrameRate = freqs[fMpgLibPrivate.fr.sampling_frequency];
		fChannelCount = fMpgLibPrivate.fr.stereo;
		fFrameSize = 2 * fChannelCount;
	}

//	printf("decode new data result %d\n", result);
		
//	printf("mp3Decoder::Decode: decoded %d bytes into %d bytes\n", chunkSize, outsize);
		
	fResidualBuffer = fDecodeBuffer;
	fResidualBytes = outsize;

	{
//		const uint16 *p = (const uint16 *)fDecodeBuffer;
//		printf("out %04x %04x %04x %04x %04x %04x %04x %04x\n", p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7]);
	}	
	return B_OK;
}


bool
AudioDecoder::IsValidStream(const uint8 *buffer, int size)
{
	// check 3 consecutive frame headers to make sure
	// that the length encoded in the header is correct,
	// and also that mpeg version and layer do not change
	int length1 = GetFrameLength(buffer);
	if (length1 < 0 || (length1 + 4) > size)
		return false;
	int version_index1 = (buffer[1] >> 3) & 0x03;
	int layer_index1 = (buffer[1] >> 1) & 0x03;
	int length2 = GetFrameLength(buffer + length1);
	if (length2 < 0 || (length1 + length2 + 4) > size)
		return false;
	int version_index2 = (buffer[length1 + 1] >> 3) & 0x03;
	int layer_index2 = (buffer[length1 + 1] >> 1) & 0x03;
	if (version_index1 != version_index2 || layer_index1 != layer_index1)
		return false;
	int length3 = GetFrameLength(buffer + length1 + length2);
	if (length3 < 0)
		return false;
	int version_index3 = (buffer[length1 + length2 + 1] >> 3) & 0x03;
	int layer_index3 = (buffer[length1 + length2 + 1] >> 1) & 0x03;
	if (version_index2 != version_index3 || layer_index2 != layer_index3)
		return false;
	return true;
}


int
AudioDecoder::GetFrameLength(const void *header)
{
	const uint8 *h = (const uint8 *)header;

	if (h[0] != 0xff)
		return -1;
	if ((h[1] & 0xe0) != 0xe0)
		return -1;
	
	int mpeg_version_index = (h[1] >> 3) & 0x03;
	int layer_index = (h[1] >> 1) & 0x03;
	int bitrate_index = (h[2] >> 4) & 0x0f;
	int sampling_rate_index = (h[2] >> 2) & 0x03;
	int padding = (h[2] >> 1) & 0x01;
	/* not interested in the other bits */
	
	int bitrate = bit_rate_table[mpeg_version_index][layer_index][bitrate_index];
	int framerate = frame_rate_table[mpeg_version_index][sampling_rate_index];
	
	if (!bitrate || !framerate)
		return -1;

	int length;	
	if (layer_index == 3) // layer 1
		length = ((144 * 1000 * bitrate) / framerate) + (padding * 4);
	else // layer 2 & 3
		length = ((144 * 1000 * bitrate) / framerate) + padding;

#if 0
	TRACE("%s %s, %s crc, bit rate %d, frame rate %d, padding %d, frame length %d\n",
		mpeg_version_index == 0 ? "mpeg 2.5" : (mpeg_version_index == 2 ? "mpeg 2" : "mpeg 1"),
		layer_index == 3 ? "layer 1" : (layer_index == 2 ? "layer 2" : "layer 3"),
		(h[1] & 0x01) ? "no" : "has",
		bitrate, framerate, padding, length);
#endif

	return length;
}


extern "C" DecoderInterface * 
instantiate_mpeg2_audio_decoder(get_next_chunk_func next_chunk, void *cookie, const char *param)
{
	return new AudioDecoder(next_chunk, cookie);
}
